'============================================================
' Barionet Extension Tool
'
' This BCL application supports the Barionet extensions R6, IO12 and the X8 (limited, digital IO only).
' For Modbus address change you have to start the page add_main.html 
' (if this cob file is loaded !) in your browser i.e. http://192.168.1.10/add_main.html
' (there is also a help page add_help.html)
' Note, all fields in the upper section must be filled!
'
' The serial number you can find on the label on the backside of your 
' Barionet extension unit (on IO12 below the barcode).
' Serial No: i.e.  1  119   ;   R272,R273 and R275 usually not used = 0   ;
' R274 = should be the modbus address ( i.e. 253) 
'
' On this page you also see an overview of all properly connected and addresses extension devices.
' This working BCL programming Demo allows the control of the digital inputs and digital output registers and the relays
' of upto 7 connected Barionet extension units ' over web, TCP, UDP,CGI, SNMP and Modbus(TCP).
' The inputs are mapped to Barionet registers 211..294 and the outputs to register 111 to 171 and the relay to registers 11 to 52.
' For UDP commands you can use the normal Barionet syntax (e.g. setio,111,1 ).
'
' The Io register numbers for each extension unit depend from the type of the connected extensions and the Modbus addressing
' and are displayed on the integrated Modbus addressing page.
' The variable "adr" defines address for the different the modbus devices.
' Currently are the addresses 255, 254, 253, 252, 251, 250 and 249 supported in this example code.

' The application was tested, but nevertheless: No performance guarantees - code use at your own risk
'=============================================================
'Release Notes
'
'version 1.00
'03.09.13 - SG - initial version for 7 Barionet extension units
'version 1.01 new
'15.08.21 - SG - speed up loop time by learned CRC
'=============================================================
DIM Com$(25)
DIM a,i,o,u,x,y,k,j,uu,cp	 		 	
DIM cgiadr,adr,mbreg,mbreg_h,mblen,mbok
DIM mlen,coil
DIM j2000,i1000,init,err,crc,tmp
DIM m$(50)
DIM mr$(50)
DIM mx$(128)
DIM _CGI_$,c$
DIM sldl					' syslog debug level
DIM p(7)					' check valiable for loop counter
DIM sn1, sn1l
DIM sn2, sn2l
DIM r272, r272l
DIM r273, r273l
DIM r274, r274l
DIM r275, r275l
DIM name$(20),msg(50),cmd$(20),p$
DIM rport, rhost$(20)
DIM start_out(7), out_all
DIM old_out(100), old_in(7), start_in(7), in_all, ios(7)
DIM start_rel(7), rel_all, old_rel(50)
DIM dev(7), fw(7), s_n1(7), s_n2(7)
DIM stor_crc(7)				' CRC storage variable

name$ = sprintf$("%1A",1)
SYSLOG "Start Barionet extension tool (fix range) with ICgraph support on "+name$ + " v1.01", 1
name$ = mid$(name$,13,3)


Open "STP:1" as 1
  READ 1,p$					' read Barionet settings
  sldl=MIDGET(p$,92,1)		' copy SYSLOG debug level from Barionet setup
  SYSLOG "SYSLOG Level = "+STR$(sldl)
  CLOSE 1


'------------------- opens communication channels ---------------------------------
com$ = "COM:19200,E,8,1,485:2"		' open rs485
Open COM$ as 4						' serial interface parameters are set in the setup !


_CGI_$ = ""
ON CGI GOSUB 50
SYSLOG "CGI handler enabled", 1  


Open "UDP:0.0.0.0:12300" as 5 		' command port for UDP from ICgraph (NB! fixed port 12300)

'------------------------------------------------------------------------------------


'------------------------------------------------------------------
2 'read initial status of Extension devices
' -----------------------------------------------------------------

init=0
FOR y=0 to 6
	p(y)=0		' reset loop counter
NEXT y
rel_all=10
out_all=110
in_all=210
FOR y = 0 TO 6
     dev(y)=0					' resets values to 0
     fw(y)=0					' resets values to 0
     s_n1(y)=0					' resets values to 0
     s_n2(y)=0					' resets values to 0
     IOCTL 761+(y*2),0			' writes to Barionet register 761 - 779
     IOCTL 762+(y*2),0					
     IOCTL 781+(y*2),0			' writes to Barionet register 781 - 799
     IOCTL 782+(y*2),0					
NEXT y


mbreg_h=1						' high register byte
mbreg=0							' reads firmware version + serial number
mblen=6
FOR y = 0 TO 6
adr=255-y
 GOSUB 200
 IF mbok=1 THEN 
   dev(y)=(256*MIDGET(mr$,4,1)) + MIDGET(mr$,4+1,1)	' reads device type
   fw(y)=(256*MIDGET(mr$,6,1)) + MIDGET(mr$,6+1,1)	' reads firmware version
   s_n1(y)=MIDGET(mr$,9,1)				' reads first part of Serial number
   s_n2(y)=MIDGET(mr$,11,1)				' reads second part of serial number   
   
   IF dev(y)=177 THEN			' 177 = 0xB1 for R6 device
     ios(y)=6 : start_rel(y)=rel_all+1 : IOCTL 761+(y*2),start_rel(y)
     rel_all=rel_all+ios(y) : IOCTL 762+(y*2),rel_all 
   ENDIF

   IF dev(y)=178 THEN 			' 178 = 0xB2 for X8 device
     ios(y)=8 : start_out(y)=out_all+1 : IOCTL 761+(y*2),start_out(y)
     out_all=out_all+ios(y) : start_in(y)=in_all+1 : IOCTL 781+(y*2),start_in(y)
     IOCTL 762+(y*2),out_all : in_all=in_all+ios(y) : IOCTL 782+(y*2),in_all
   ENDIF

   IF dev(y)=179 THEN			' 179 = 0xB3 for IO12 device
     ios(y)=12 : start_out(y)=out_all+1 : IOCTL 761+(y*2),start_out(y)
     out_all=out_all+ios(y) : start_in(y)=in_all+1 : IOCTL 781+(y*2),start_in(y)
     IOCTL 762+(y*2),out_all : in_all=in_all+ios(y) : IOCTL 782+(y*2),in_all
   ENDIF
 ENDIF
NEXT y


'--------------read initial output values-------------------------------
5
FOR y = 0 TO 6
IF dev(y)=179 THEN				' 179 = 0xB3 for IO12 device
  mbreg=0						' Modbus register for IO12 outputs
  mbreg_h=0
  mblen=2
  adr=255-y
  GOSUB 200						' send modbus read command and read response, will return buffer and 
  IF mbok=1 THEN
	u = MIDGET(mr$,4,-2)  		' get byte from IO12	
	FOR i = 0 TO 11
	  IOCTL start_out(y)+i,1+SHR(AND(u,2^i),i)-1  'copies output registers to Barionet reg 1xx (here 1 for close, 0 for open)
	  SYSLOG "IO12 ("+SPRINTF$("%u",adr)+") Output "+SPRINTF$("%u",1+i)+" = "+SPRINTF$("%u",IOSTATE(start_out(y)+i)), 1
	  old_out(start_out(y)-110+i)=IOSTATE(start_out(y)+i)
	NEXT i
  ENDIF
ENDIF
IF dev(y)=178 THEN				' 178 = 0xB2 for X8 device
	 mbreg=0	
	 mbreg_h=0					' Modbus register for outputs
	 mblen=1					' numbers of registers to read
	 adr=255-y
	 GOSUB 200 					' send modbus read command, will return buffer and mbok=1
	   IF mbok=1 THEN
      	      u = MIDGET(m$,4,-2)		' get WORD from X8
	      FOR i = 0 TO 7
        	IOCTL start_out(y)+i,1+shr(AND(u,2^i),i)-1 'copies output registers 1xx (here 1 for close, 0 for open)
        	SYSLOG "X8 ("+SPRINTF$("%u",adr)+") Output "+SPRINTF$("%u",1+i)+" = "+SPRINTF$("%u",IOSTATE(start_out(y)+i)), 1
		old_out(start_out(y)-110+i)=IOSTATE(start_out(y)+i)
      	      NEXT i
	   ENDIF
ENDIF
IF dev(y) = 177 THEN 			' 177 = 0xB1 for R6 device
	 mbreg=0	
	 mbreg_h=0					' Modbus register for outputs
	 mblen=1					' numbers of registers to read
	 adr=255-y
	 GOSUB 200 					' send modbus read command, will return buffer and mbok=1
	   IF mbok=1 THEN
		  u = MIDGET(mr$,4,-2)  ' get byte from R6
  		    FOR i = 0 TO 5
      		      IOCTL start_rel(y)+i,SHR(AND(u,2^i),i) 	'copies output registers to Barionet reg 1x (here 1 for close, 0 for open)
       		      SYSLOG "R6 ("+SPRINTF$("%u",adr)+") relay "+SPRINTF$("%u",1+i)+" = "+SPRINTF$("%u",IOSTATE(start_rel(y)+i)), 1
		      old_rel(start_rel(y)-10+i)=IOSTATE(start_rel(y)+i)
		    NEXT i
	   ENDIF
ENDIF
NEXT y

'--------------------------------------------------------------------------------------
mbreg_h=0
mblen=2
mbreg=1				' Modbus register for inputs

'-------------------- start main loop ------------------------------------------------
10
FOR y = 0 TO 6
  adr=255-y									' set Modbus address for selelcted R6
  IF dev(y) = 177 THEN     					' when R6
	FOR c = 0 TO 5							' defining Barionet register range for the selected R6
	  ioval=IOSTATE(start_rel(y)+c)				
	  IF ioval<>old_rel(start_rel(y)-10+c) THEN	' testing for relay (R6) register change
		coil=c			
		IF ioval=1 THEN
          	    GOSUB 900     				' set relay to 1
	  		ELSE
            		GOSUB 800     			' set relay to 0
       		ENDIF
		old_rel(start_rel(y)-10+c)=ioval	' set new IO state to old state
    	  ENDIF
	  NEXT c
  ENDIF
  
  IF OR(dev(y)=178,dev(y)=179) THEN			' testing for output (X8 or IO12) register change
    FOR c = 0 TO ios(y)-1					' defining Barionet output range for the selected X8/IO12
	  ioval=IOSTATE(start_out(y)+c)
	  IF ioval<>old_out(start_out(y)-110+c) THEN
          	coil=c
		IF ioval=1 THEN
          	    GOSUB 900     				' set output to 1
	  		ELSE
            		GOSUB 800     			' set output to 0
       		ENDIF
		old_out(start_out(y)-110+c)=ioval	' set new IO state to old state
    	  ENDIF
    NEXT c
    
'---------------section for input check--------------------------------------------------
    GOSUB 200					        ' read input state
		p(y)=p(y)+1
    IF mbok=1 THEN					' when something received
      IF old_in(y) <> MIDGET(mr$,4,-2) THEN old_in(y) = MIDGET(mr$,4,-2) ELSE GOTO 15 ' compare new X8 state with older state
      IF dev(y) = 178 THEN              		' when X8 then
	 FOR i = 0 TO 7					' check 8 registers for X8
	   u = 1+shr(and(old_in(y),2^i),i)-1		' extract input register
	   IF u=1 THEN IOCTL start_in(y)+i,0 ELSE IOCTL start_in(y)+i,1	' invert value and write to Barionet register (digital input)	
           SYSLOG "X8 ("+STR$(adr)+") Input "+SPRINTF$("%u",i+1)+" = "+SPRINTF$("%u",IOSTATE(start_in(y)+i)), 1
         NEXT i
         ELSE
      	    FOR i = 0 TO 11				' for IO12 check 12 registers
              IOCTL start_in(y)+i, 1 + SHR(AND(old_in(y),2^i), i)-1 'copies input registers 2xx (here 1 for close, 0 for open)
              SYSLOG "IO12 ("+STR$(adr)+") Input "+SPRINTF$("%u",i+1)+" = "+SPRINTF$("%u",IOSTATE(start_in(y)+i)), 1
            NEXT i
	ENDIF
    15
    ENDIF
  ENDIF
NEXT y


GOSUB 20
DELAY 0
IF init=1 THEN
   GOTO 2						' initialize again when Modbus parameters are changed
     ELSE 
     GOTO 10					' otherwise perform normal main loop
ENDIF
END								' end main loop

'------------------------------------------------------------------------------

20 'checks received command from UDP port 12300 (ICgraph support)
'------------------------------------------------------------------------------
    IF LASTLEN(5) < 0 THEN
      SYSLOG " Incoming Command" , 3
      rport=RMTPORT(5)
      rhost$=RMTHOST$(5)
      SYSLOG "answer to Remote port : " + STR$(rport), 3
      SYSLOG "answer to Remote IP address : " + rhost$, 3
      READ 5,cmd$ : msg$="wrong command: "+cmd$+"\r"
      IF cmd$ = "c=65535" THEN msg$="<BARIONET><NAME>"+name$+"</NAME></BARIONET>" ' broadcast DISCOVER
      IF MID$(cmd$,1,6) = "setio," THEN
         cp = INSTR(7,cmd$,",")        	 		' comma pos
         uu = VAL(MID$(cmd$,7,cp-7))   	 		' get register from request
         IOCTL uu,VAL(MID$(cmd$,cp+1)) 	 		' set register
         IF err THEN msg$="state,err\r" ELSE msg$="state,"+STR$(uu)+","+STR$(IOSTATE(uu))+"\r" ' set reply
      ENDIF
      IF MID$(cmd$,1,6) = "getio," THEN
         uu = val(mid$(cmd$,7))
         IF err THEN msg$="state,err\r" ELSE msg$="state,"+STR$(uu)+","+STR$(IOSTATE(uu))+"\r" ' set reply
      ENDIF
      WRITE 5,msg$,0,rhost$,rport 	 			' send reply
      SYSLOG msg$, 3
    ENDIF
    RETURN

'------------------------------------------------------------------------
50 'CGI Handler to change Modbus address settings (from Modbus Addressing tool)
'------------------------------------------------------------------------
  c$ = _CGI_$
  _CGI_$ = ""
  WRITE -1,"HTTP/1.1 200 OK\r\n\r\n",0
   _CGI_$ = "*"
  IF LEN(c$) > 0 THEN
    SYSLOG "CGI : " + c$, 1
    sn1 = INSTR(1, c$, "sn1=")
    sn2 = INSTR(1, c$, "sn2=")
    r272 = INSTR(1, c$, "r272=")
    r273 = INSTR(1, c$, "r273=")
    r274 = INSTR(1, c$, "r274=")
    r275 = INSTR(1, c$, "r275=")
    IF AND(sn1 > 0, sn2 > 0, r272 > 0, r273 > 0, r274 > 0, r275 > 0) THEN
      sn1 = sn1 + 4
      sn2 = sn2 + 4
      r272 = r272 + 5
      r273 = r273 + 5
      r274 = r274 + 5
      r275 = r275 + 5
      tmp = INSTR(sn1, c$, "&")
      IF tmp > 0 THEN
        sn1l = tmp - sn1
      ELSE
        sn1l = LEN(c$) - sn1
      ENDIF
      tmp = INSTR(sn2, c$, "&")
      IF tmp > 0 THEN
        sn2l = tmp - sn2
      ELSE
        sn2l = LEN(c$) - sn2
      ENDIF
      tmp = INSTR(r272, c$, "&")
      IF tmp > 0 THEN
        r272l = tmp - r272
      ELSE
        r272l = LEN(c$) - r272
      ENDIF
      tmp = INSTR(r273, c$, "&")
      IF tmp > 0 THEN
        r273l = tmp - r273
      ELSE
        r273l = LEN(c$) - r273
      ENDIF
      tmp = INSTR(r274, c$, "&")
      IF tmp > 0 THEN
        r274l = tmp - r274
      ELSE
        r274l = LEN(c$) - r274 + 1
      ENDIF
      tmp = INSTR(r275, c$, "&")
      IF tmp > 0 THEN
        r275l = tmp - r275
      ELSE
        r275l = LEN(c$) - r275 + 1
      ENDIF
      IF AND(sn1l > 0, sn2l > 0, r272l > 0, r273l > 0, r274l > 0, r275l > 0) THEN
        sn1  = VAL(MID$(c$, sn1, sn1l))
        sn2  = VAL(MID$(c$, sn2, sn2l))
        r272 = VAL(MID$(c$, r272, r272l))
        r273 = VAL(MID$(c$, r273, r273l))
        r274 = VAL(MID$(c$, r274, r274l))
        r275 = VAL(MID$(c$, r275, r275l))
        MIDSET m$, 1, 1, 0     				' Broadcast
        MIDSET m$, 2, 1, 69    				' Set config
        MIDSET m$, 3, 1, sn1   				' Serial number, part 1
        MIDSET m$, 4, 1, sn2   				' Serial number, part 2        
        MIDSET m$, 5, 1, sn1   				' Serial number, part 1 again
        MIDSET m$, 6, 1, sn2   				' Serial number, part 2 again
        MIDSET m$, 7, 1, r272  				' Relay power on defaults
        MIDSET m$, 8, 1, r273  				' Serial interface mode
        MIDSET m$, 9, 1, r274  				' Modbus address
        MIDSET m$, 10, 1, r275 				' Reserved  
        mlen = 12
	GOSUB 1001
	GOSUB 700  
	WRITE 4, m$, mlen
	GOSUB 2000		 						' will SYSLOG Modbus command for debugging
	GOSUB 700
	SYSLOG "Set config for device " + STR$(sn1) + "-" + STR$(sn2), 1
	ELSE
	  SYSLOG "Invalid CGI request (Missing Parameter Values)", 1
	ENDIF
      ELSE
	tmp = INSTR(1, c$, "cgiadr=")			' command will be used for future features
	IF AND(tmp > 0, LEN(c$) > 4) THEN
	  tmp = tmp + 4
          tmp = VAL(MID$(c$, tmp, LEN(c$) - tmp + 1))
          cgiadr = tmp
          ELSE
            SYSLOG "Invalid CGI request (Missing Parameters Name)", 1
	ENDIF    
      ENDIF
  ELSE
    SYSLOG "Invalid CGI request (No Command)", 1
ENDIF
init=1
FOR y=0 to 6
	p(y)=0		' reset loop counter
NEXT y
RETURN
	
'-------------------------------------------------------------------------------
200 'modbus read for intial relay state and for input state
'-------------------------------------------------------------------------------
  MIDSET m$,1,1,adr			' set address
  MIDSET m$,2,1,3			' read registers command
  MIDSET m$,3,1,mbreg_h		' hi address
  MIDSET m$,4,1,mbreg		' low address
  MIDSET m$,5,1,0			' hi number
  MIDSET m$,6,1,mblen		' lo number
  'MIDSET m$,7,1,0			' crc value
  'MIDSET m$,8,1,0			' crc value
  mlen=8
  IF p(y)>1 THEN								' check for second or higher loop, then we should have copied CRC already
	MIDSET m$, (mlen - 1), 2, stor_crc(y)		' using copied CRC to speed up loop time
	IF sldl>4 then SYSLOG "used copied CRC",5	' checking SYSLOG debug level 
	p(y)=2
	ELSE
		GOSUB 1000			' else calculate modbus crc
  ENDIF
  GOSUB 700 				' read out stuff from interface, if line noise (unlikely)
  WRITE 4,m$,mlen			' send out the block
  GOSUB 2000
  GOSUB 300					' wait for answer
DELAY 1
RETURN


'========================================================================
' Modbus read - read back answer from modbus in mr$, return status with it
'========================================================================
300
j2000=_TMR_(0)

400
READ 4,mr$,10		 		' try to read
IF AND(_TMR_(0)-j2000<100,LASTLEN(4)<=0) THEN GOTO 400 'wait upto 100 ms for response
mlen=LASTLEN(4)
IF mlen>3 THEN
	a=1						' set marker for debugging info
	GOSUB 2000				' send debugging SYSLOG
	crc=&HFFFF				' start of CRC check for incoming messages
	FOR i1000=1 TO mlen-2
		crc=xor(crc,MIDGET(mr$,i1000,1))
		IF AND(crc,1) THEN crc=xor(crc,&H14002)
		IF AND(crc,2) THEN crc=xor(crc,&H28004)
		IF AND(crc,4) THEN crc=xor(crc,&H50008)
		IF AND(crc,8) THEN crc=xor(crc,&HA0010)
		IF AND(crc,16) THEN crc=xor(crc,&H140020)
		IF AND(crc,32) THEN crc=xor(crc,&H280040)
		IF AND(crc,64) THEN crc=xor(crc,&H500080)
		IF AND(crc,128) THEN crc=xor(crc,&HA00100)
		crc=SHR(crc,8)
	NEXT i1000				' end of CRC check for incoming messages 
	IF MIDGET(mr$,mlen-1,2)<>crc THEN
		SYSLOG "read: crc error"
		mbok=-2
		ELSE
           mbok=1
	ENDIF
	ELSE
	SYSLOG "no answer from modbus device " + STR$(adr), 7
	mbok=-1
ENDIF
RETURN

---------------------------------------------------------
'read/clean noise trash from the line

700 							' Read response
READ 4, mx$,1		 			' try to read
RETURN

'=========================================================
'section for outputs
'=====================

800 'Reset Coil
  MIDSET m$, 1, 1, adr  		' Modbus address
  MIDSET m$, 2, 1, 5    		' Single Coil
  MIDSET m$, 3, 1, 0    		' Address Hi
  MIDSET m$, 4, 1, coil 		' Address Lo
  MIDSET m$, 5, 1, 0    		' Value Hi
  MIDSET m$, 6, 1, 0    		' Value Lo
  mlen = 8
  GOSUB 1001
  GOSUB 700
  WRITE 4, m$, mlen
  GOSUB 2000
  SYSLOG "Reset coil " + STR$(coil) + " on device " + STR$(adr) + " issued", 5
RETURN

900 'Set Coil
  MIDSET m$, 1, 1, adr  		' Modbus address
  MIDSET m$, 2, 1, 5    		' Single Coil
  MIDSET m$, 3, 1, 0    		' Address Hi
  MIDSET m$, 4, 1, coil 		' Address Lo
  MIDSET m$, 5, 1, 255  		' Value Hi
  MIDSET m$, 6, 1, 255  		' Value Lo
  mlen = 8
  GOSUB 1001
  GOSUB 700
  WRITE 4, m$, mlen
  GOSUB 2000
  SYSLOG "Set coil " + STR$(coil) + " on device " + STR$(adr) + " issued", 5
RETURN

'==============================================================
1000 'Modbus CRC calc for same Modbus requests (checking inputs) with CRC learning
'==============================================================
  crc = &HFFFF
  FOR i1000 = 1 TO (mlen - 2)
    crc = XOR(crc, MIDGET(m$, i1000, 1))
    FOR j = 1 TO 8
      k = AND(crc, 1)
      crc = SHR(crc, 1)
      IF k = 1 THEN crc = XOR(crc, &HA001)
    NEXT j
  NEXT i1000        
  MIDSET m$, (mlen - 1), 2, crc
  stor_crc(y)=crc				' learning crc for faster loop time
RETURN

'===========================================================
1001 'Modbus CRC calc for changing Modbus messages (e.g. for output control)
'===========================================================
  crc = &HFFFF
  FOR i1000 = 1 TO (mlen - 2)
    crc = XOR(crc, MIDGET(m$, i1000, 1))
    FOR j = 1 TO 8
      k = AND(crc, 1)
      crc = SHR(crc, 1)
      IF k = 1 THEN crc = XOR(crc, &HA001)
    NEXT j
  NEXT i1000        
  MIDSET m$, (mlen - 1), 2, crc
RETURN

'------------------------------------------------------------------------
2000 ' SYSLOG DEBUGGING for MODBUS communication
' to speed up the loop time this routine is only used if SYSLOG level is 6 or higher in Barionet setup
'------------------------------------------------------------------------
If sldl>5 then						' checking SYSLOG debug level 
	IF a=1 THEN						' a=1 is "read" , a=0 is "write"
		p$="read "+STR$(adr)+SPRINTF$(": %d bytes: ",mlen)	' prepare IO12/X8 response for SYSLOG
		FOR o=1 TO mlen 
			p$=p$+SPRINTF$(" %02x",MIDGET(mr$,o,1)) 	' prepare IO12/X8 response/command for SYSLOG
		NEXT o
		ELSE
			p$="write "+STR$(adr)+SPRINTF$(": %d bytes: ",mlen)	' prepare IO12/X8/R6 response for SYSLOG
			FOR o=1 TO mlen 
				p$=p$+SPRINTF$(" %02x",MIDGET(m$,o,1)) 	' prepare IO12/X8/R6 response/command for SYSLOG
			NEXT o
	ENDIF		
	SYSLOG p$, 6					' syslog IO12/X8 Modbus response string
	a=0
ENDIF
RETURN
